@@ -1,4 +1,4 @@
-/* $Header: /src/pub/tcsh/tc.who.c,v 3.44 2005/03/03 23:44:45 kim Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/tc.who.c,v 3.55 2012/01/10 17:07:51 christos Exp $ */
 /*
  * tc.who.c: Watch logins and logouts...
  */
@@ -32,7 +32,7 @@
  */
 #include "sh.h"
 
-RCSID("$Id: tc.who.c,v 3.44 2005/03/03 23:44:45 kim Exp $")
+RCSID("$tcsh: tc.who.c,v 3.55 2012/01/10 17:07:51 christos Exp $")
 
 #include "tc.h"
 
@@ -44,6 +44,11 @@ RCSID("$Id: tc.who.c,v 3.44 2005/03/03 23:44:45 kim Exp $")
 
 #ifdef HAVE_UTMPX_H
 # include <utmpx.h>
+# define UTNAMLEN	sizeof(((struct utmpx *) 0)->ut_name)
+# define UTLINLEN	sizeof(((struct utmpx *) 0)->ut_line)
+# ifdef HAVE_STRUCT_UTMPX_UT_HOST
+#  define UTHOSTLEN	sizeof(((struct utmpx *) 0)->ut_host)
+# endif
 /* I just redefine a few words here.  Changing every occurrence below
  * seems like too much of work.  All UTMP functions have equivalent
  * UTMPX counterparts, so they can be added all here when needed.
@@ -55,57 +60,69 @@ RCSID("$Id: tc.who.c,v 3.44 2005/03/03 23:44:45 kim Exp $")
 #  define TCSH_PATH_UTMP _PATH_UTMPX
 # elif defined(UTMPX_FILE)
 #  define TCSH_PATH_UTMP UTMPX_FILE
-# endif /* __UTMPX_FILE && !UTMPX_FILE */
-# ifdef TCSH_PATH_UTMP
+# elif __FreeBSD_version >= 900000
+#  /* Why isn't this defined somewhere? */
+#  define TCSH_PATH_UTMP "/var/run/utx.active"
+# elif defined(__hpux)
+#  define TCSH_PATH_UTMP "/etc/utmpx"
+# endif
+# if defined(TCSH_PATH_UTMP) || !defined(HAVE_UTMP_H)
 #  define utmp utmpx
-#  if defined(HAVE_STRUCT_UTMP_UT_TV)
-#   define ut_time ut_tv.tv_sec
-#  elif defined(HAVE_STRUCT_UTMP_UT_XTIME)
-#   define ut_time ut_xtime
-#  endif
-#  ifdef HAVE_STRUCT_UTMP_UT_USER
-#   define ut_name ut_user
-#  endif
-#  ifdef HAVE_GETUTENT
+#  define TCSH_USE_UTMPX
+#  if defined(HAVE_GETUTENT) || defined(HAVE_GETUTXENT)
 #   define getutent getutxent
 #   define setutent setutxent
 #   define endutent endutxent
-#  endif /* HAVE_GETUTENT */
-# else
-#  ifdef HAVE_UTMP_H
-#   include <utmp.h>
-#  endif /* WINNT_NATIVE */
-# endif /* TCSH_PATH_UTMP */
-#else /* !HAVE_UTMPX_H */
-# ifdef HAVE_UTMP_H
-#  include <utmp.h>
-# endif /* WINNT_NATIVE */
+#  endif /* HAVE_GETUTENT || HAVE_GETUTXENT */
+#  if defined(HAVE_STRUCT_UTMPX_UT_TV)
+#   define ut_time ut_tv.tv_sec
+#  elif defined(HAVE_STRUCT_UTMPX_UT_XTIME)
+#   define ut_time ut_xtime
+#  endif
+#  if defined(HAVE_STRUCT_UTMPX_UT_USER)
+#   define ut_name ut_user
+#  endif
+# endif /* TCSH_PATH_UTMP || !HAVE_UTMP_H */
 #endif /* HAVE_UTMPX_H */
 
-#ifndef BROKEN_CC
-# define UTNAMLEN	sizeof(((struct utmp *) 0)->ut_name)
-# define UTLINLEN	sizeof(((struct utmp *) 0)->ut_line)
-# ifdef HAVE_STRUCT_UTMP_UT_HOST
-#  ifdef _SEQUENT_
-#   define UTHOSTLEN	100
-#  else
-#   define UTHOSTLEN	sizeof(((struct utmp *) 0)->ut_host)
-#  endif
-# endif	/* HAVE_STRUCT_UTMP_UT_HOST */
-#else
+#if !defined(TCSH_USE_UTMPX) && defined(HAVE_UTMP_X)
+# include <utmp.h>
+# if defined(HAVE_STRUCT_UTMP_UT_TV)
+#  define ut_time ut_tv.tv_sec
+# elif defined(HAVE_STRUCT_UTMP_UT_XTIME)
+#  define ut_time ut_xtime
+# endif
+# if defined(HAVE_STRUCT_UTMP_UT_USER)
+#  define ut_name ut_user
+# endif
+#endif /* !TCSH_USE_UTMPX && HAVE_UTMP_H */
+
+#if defined(HAVE_UTMP_H) && !defined(HAVE_UTMPX_H)
+# ifndef BROKEN_CC
+#  define UTNAMLEN	sizeof(((struct utmp *) 0)->ut_name)
+#  define UTLINLEN	sizeof(((struct utmp *) 0)->ut_line)
+#  ifdef HAVE_STRUCT_UTMP_UT_HOST
+#   ifdef _SEQUENT_
+#    define UTHOSTLEN	100
+#   else
+#    define UTHOSTLEN	sizeof(((struct utmp *) 0)->ut_host)
+#   endif
+#  endif	/* HAVE_STRUCT_UTMP_UT_HOST */
+# else
 /* give poor cc a little help if it needs it */
 struct utmp __ut;
 
-# define UTNAMLEN	sizeof(__ut.ut_name)
-# define UTLINLEN	sizeof(__ut.ut_line)
-# ifdef HAVE_STRUCT_UTMP_UT_HOST
-#  ifdef _SEQUENT_
-#   define UTHOSTLEN	100
-#  else
-#   define UTHOSTLEN	sizeof(__ut.ut_host)
-#  endif
-# endif /* HAVE_STRUCT_UTMP_UT_HOST */
+#  define UTNAMLEN	sizeof(__ut.ut_name)
+#  define UTLINLEN	sizeof(__ut.ut_line)
+#  ifdef HAVE_STRUCT_UTMP_UT_HOST
+#   ifdef _SEQUENT_
+#    define UTHOSTLEN	100
+#   else
+#    define UTHOSTLEN	sizeof(__ut.ut_host)
+#   endif
+#  endif /* HAVE_STRUCT_UTMP_UT_HOST */
 #endif /* BROKEN_CC */
+#endif /* HAVE_UTMP_H */
 
 #ifndef TCSH_PATH_UTMP
 # ifdef	UTMP_FILE
@@ -124,9 +141,9 @@ struct who {
     char    who_name[UTNAMLEN + 1];
     char    who_new[UTNAMLEN + 1];
     char    who_tty[UTLINLEN + 1];
-#ifdef HAVE_STRUCT_UTMP_UT_HOST
+#ifdef UTHOSTLEN
     char    who_host[UTHOSTLEN + 1];
-#endif /* HAVE_STRUCT_UTMP_UT_HOST */
+#endif /* UTHOSTLEN */
     time_t  who_time;
     int     who_status;
 };
@@ -135,9 +152,9 @@ static struct who whohead, whotail;
 static time_t watch_period = 0;
 static time_t stlast = 0;
 #ifdef WHODEBUG
-static	void	debugwholist	__P((struct who *, struct who *));
+static	void	debugwholist	(struct who *, struct who *);
 #endif
-static	void	print_who	__P((struct who *));
+static	void	print_who	(struct who *);
 
 
 #define ONLINE		01
@@ -145,6 +162,7 @@ static	void	print_who	__P((struct who *));
 #define CHANGED		04
 #define STMASK		07
 #define ANNOUNCE	010
+#define CLEARED		020
 
 /*
  * Karl Kleinpaste, 26 Jan 1984.
@@ -153,7 +171,7 @@ static	void	print_who	__P((struct who *));
  * when doing pointer-chase searches.
  */
 void
-initwatch()
+initwatch(void)
 {
     whohead.who_next = &whotail;
     whotail.who_prev = &whohead;
@@ -164,7 +182,7 @@ initwatch()
 }
 
 void
-resetwatch()
+resetwatch(void)
 {
     watch_period = 0;
     stlast = 0;
@@ -175,19 +193,15 @@ resetwatch()
  * Watch /etc/utmp for login/logout changes.
  */
 void
-watch_login(force)
-    int force;
+watch_login(int force)
 {
     int     comp = -1, alldone;
     int	    firsttime = stlast == 1;
-#ifdef HAVE_GETUTENT
+#if defined(HAVE_GETUTENT) || defined(HAVE_GETUTXENT)
     struct utmp *uptr;
 #else
     int utmpfd;
 #endif
-#ifdef BSDSIGS
-    sigmask_t omask;
-#endif				/* BSDSIGS */
     struct utmp utmp;
     struct who *wp, *wpnew;
     struct varent *v;
@@ -206,19 +220,12 @@ watch_login(force)
 #endif /* WINNT_NATIVE */
 
     /* stop SIGINT, lest our login list get trashed. */
-#ifdef BSDSIGS
-    omask = sigblock(sigmask(SIGINT));
-#else
-    (void) sighold(SIGINT);
-#endif
+    pintr_disabled++;
+    cleanup_push(&pintr_disabled, disabled_cleanup);
 
     v = adrof(STRwatch);
     if ((v == NULL || v->vec == NULL) && !force) {
-#ifdef BSDSIGS
-	(void) sigsetmask(omask);
-#else
-	(void) sigrelse(SIGINT);
-#endif
+	cleanup_until(&pintr_disabled);
 	return;			/* no names to watch */
     }
     if (!force) {
@@ -236,7 +243,7 @@ watch_login(force)
 	 * before we are due -amol 6/5/97
 	 */
 	if (!ncbs_posted) {
-	    unsigned long tdiff = t - watch_period;
+	    time_t tdiff = t - watch_period;
 	    if (!watch_period || ((tdiff  > 0) && (tdiff > (interval - 90)))) {
 		start_ncbs(vp);
  		ncbs_posted = 1;
@@ -244,11 +251,7 @@ watch_login(force)
 	}
 #endif /* WINNT_NATIVE */
     if (t - watch_period < interval) {
-#ifdef BSDSIGS
-	(void) sigsetmask(omask);
-#else
-	(void) sigrelse(SIGINT);
-#endif
+	cleanup_until(&pintr_disabled);
 	return;			/* not long enough yet... */
     }
     watch_period = t;
@@ -265,57 +268,44 @@ watch_login(force)
 	    xprintf(CGETS(26, 1,
 			  "cannot stat %s.  Please \"unset watch\".\n"),
 		    TCSH_PATH_UTMP);
-# ifdef BSDSIGS
-	(void) sigsetmask(omask);
-# else
-	(void) sigrelse(SIGINT);
-# endif
+	cleanup_until(&pintr_disabled);
 	return;
     }
     if (stlast == sta.st_mtime) {
-# ifdef BSDSIGS
-	(void) sigsetmask(omask);
-# else
-	(void) sigrelse(SIGINT);
-# endif
+	cleanup_until(&pintr_disabled);
 	return;
     }
     stlast = sta.st_mtime;
-#ifdef HAVE_GETUTENT
+#if defined(HAVE_GETUTENT) || defined(HAVE_GETUTXENT)
     setutent();
 #else
-    if ((utmpfd = open(TCSH_PATH_UTMP, O_RDONLY|O_LARGEFILE)) < 0) {
+    if ((utmpfd = xopen(TCSH_PATH_UTMP, O_RDONLY|O_LARGEFILE)) < 0) {
 	if (!force)
 	    xprintf(CGETS(26, 2,
 			  "%s cannot be opened.  Please \"unset watch\".\n"),
 		    TCSH_PATH_UTMP);
-# ifdef BSDSIGS
-	(void) sigsetmask(omask);
-# else
-	(void) sigrelse(SIGINT);
-# endif
+	cleanup_until(&pintr_disabled);
 	return;
     }
+    cleanup_push(&utmpfd, open_cleanup);
 #endif
 
     /*
      * xterm clears the entire utmp entry - mark everyone on the status list
      * OFFLINE or we won't notice X "logouts"
      */
-    for (wp = whohead.who_next; wp->who_next != NULL; wp = wp->who_next) {
-	wp->who_status = OFFLINE;
-	wp->who_time = 0;
-    }
+    for (wp = whohead.who_next; wp->who_next != NULL; wp = wp->who_next)
+	wp->who_status = OFFLINE | CLEARED;
 
     /*
      * Read in the utmp file, sort the entries, and update existing entries or
      * add new entries to the status list.
      */
-#ifdef HAVE_GETUTENT
+#if defined(HAVE_GETUTENT) || defined(HAVE_GETUTXENT)
     while ((uptr = getutent()) != NULL) {
         memcpy(&utmp, uptr, sizeof (utmp));
 #else
-    while (read(utmpfd, (char *) &utmp, sizeof utmp) == sizeof utmp) {
+    while (xread(utmpfd, &utmp, sizeof utmp) == sizeof utmp) {
 #endif
 
 # ifdef DEAD_PROCESS
@@ -340,6 +330,8 @@ watch_login(force)
 	    wp = wp->who_next;/* find that tty! */
 
 	if (wp->who_next && comp == 0) {	/* found the tty... */
+	    if (utmp.ut_time < wp->who_time)
+	        continue;
 # ifdef DEAD_PROCESS
 	    if (utmp.ut_type == DEAD_PROCESS) {
 		wp->who_time = utmp.ut_time;
@@ -354,11 +346,11 @@ watch_login(force)
 	    else if (strncmp(utmp.ut_name, wp->who_name, UTNAMLEN) == 0) {
 		/* someone is logged in */ 
 		wp->who_time = utmp.ut_time;
-		wp->who_status = 0;	/* same guy */
+		wp->who_status = ONLINE | ANNOUNCE;	/* same guy */
 	    }
 	    else {
 		(void) strncpy(wp->who_new, utmp.ut_name, UTNAMLEN);
-# ifdef HAVE_STRUCT_UTMP_UT_HOST
+# ifdef UTHOSTLEN
 #  ifdef _SEQUENT_
 		host = ut_find_host(wp->who_tty);
 		if (host)
@@ -368,7 +360,7 @@ watch_login(force)
 #  else
 		(void) strncpy(wp->who_host, utmp.ut_host, UTHOSTLEN);
 #  endif
-# endif /* HAVE_STRUCT_UTMP_UT_HOST */
+# endif /* UTHOSTLEN */
 		wp->who_time = utmp.ut_time;
 		if (wp->who_name[0] == '\0')
 		    wp->who_status = ONLINE;
@@ -377,9 +369,9 @@ watch_login(force)
 	    }
 	}
 	else {		/* new tty in utmp */
-	    wpnew = (struct who *) xcalloc(1, sizeof *wpnew);
+	    wpnew = xcalloc(1, sizeof *wpnew);
 	    (void) strncpy(wpnew->who_tty, utmp.ut_line, UTLINLEN);
-# ifdef HAVE_STRUCT_UTMP_UT_HOST
+# ifdef UTHOSTLEN
 #  ifdef _SEQUENT_
 	    host = ut_find_host(wpnew->who_tty);
 	    if (host)
@@ -389,7 +381,7 @@ watch_login(force)
 #  else
 	    (void) strncpy(wpnew->who_host, utmp.ut_host, UTHOSTLEN);
 #  endif
-# endif /* HAVE_STRUCT_UTMP_UT_HOST */
+# endif /* UTHOSTLEN */
 	    wpnew->who_time = utmp.ut_time;
 # ifdef DEAD_PROCESS
 	    if (utmp.ut_type == DEAD_PROCESS)
@@ -412,18 +404,17 @@ watch_login(force)
 	    wp->who_prev = wpnew;	/* linked in now */
 	}
     }
-#ifdef HAVE_GETUTENT
+#if defined(HAVE_GETUTENT) || defined(HAVE_GETUTXENT)
     endutent();
 #else
-    (void) close(utmpfd);
+    cleanup_until(&utmpfd);
 #endif
-# if defined(HAVE_STRUCT_UTMP_UT_HOST) && defined(_SEQUENT_)
-    endutent();
-# endif
 #endif /* !WINNT_NATIVE */
 
-    if (force || vp == NULL)
+    if (force || vp == NULL) {
+	cleanup_until(&pintr_disabled);
 	return;
+    }
 
     /*
      * The state of all logins is now known, so we can search the user's list
@@ -447,8 +438,10 @@ watch_login(force)
 	    /* already printed or not right one to print */
 
 
-	    if (wp->who_time == 0)/* utmp entry was cleared */
+	    if (wp->who_status & CLEARED) {/* utmp entry was cleared */
 		wp->who_time = watch_period;
+		wp->who_status &= ~CLEARED;
+	    }
 
 	    if ((wp->who_status & OFFLINE) &&
 		(wp->who_name[0] != '\0')) {
@@ -474,17 +467,12 @@ watch_login(force)
 	    }
 	}
     }
-#ifdef BSDSIGS
-    (void) sigsetmask(omask);
-#else
-    (void) sigrelse(SIGINT);
-#endif
+    cleanup_until(&pintr_disabled);
 }
 
 #ifdef WHODEBUG
 static void
-debugwholist(new, wp)
-    struct who *new, *wp;
+debugwholist(struct who *new, struct who *wp)
 {
     struct who *a;
 
@@ -518,49 +506,47 @@ debugwholist(new, wp)
 
 
 static void
-print_who(wp)
-    struct who *wp;
+print_who(struct who *wp)
 {
-#ifdef HAVE_STRUCT_UTMP_UT_HOST
+#ifdef UTHOSTLEN
     Char   *cp = str2short(CGETS(26, 7, "%n has %a %l from %m."));
 #else
     Char   *cp = str2short(CGETS(26, 8, "%n has %a %l."));
-#endif /* HAVE_STRUCT_UTMP_UT_HOST */
+#endif /* UTHOSTLEN */
     struct varent *vp = adrof(STRwho);
-    Char buf[BUFSIZE];
+    Char *str;
 
     if (vp && vp->vec && vp->vec[0])
 	cp = vp->vec[0];
 
-    tprintf(FMT_WHO, buf, cp, BUFSIZE, NULL, wp->who_time, (ptr_t) wp);
-    for (cp = buf; *cp;)
+    str = tprintf(FMT_WHO, cp, NULL, wp->who_time, wp);
+    cleanup_push(str, xfree);
+    for (cp = str; *cp;)
 	xputwchar(*cp++);
+    cleanup_until(str);
     xputchar('\n');
 } /* end print_who */
 
 
-const char *
-who_info(ptr, c, wbuf, wbufsiz)
-    ptr_t ptr;
-    int c;
-    char *wbuf;
-    size_t wbufsiz;
+char *
+who_info(ptr_t ptr, int c)
 {
-    struct who *wp = (struct who *) ptr;
-#ifdef HAVE_STRUCT_UTMP_UT_HOST
-    char *wb = wbuf;
+    struct who *wp = ptr;
+    char *wbuf;
+#ifdef UTHOSTLEN
+    char *wb;
     int flg;
     char *pb;
-#endif /* HAVE_STRUCT_UTMP_UT_HOST */
+#endif /* UTHOSTLEN */
 
     switch (c) {
     case 'n':		/* user name */
 	switch (wp->who_status & STMASK) {
 	case ONLINE:
 	case CHANGED:
-	    return wp->who_new;
+	    return strsave(wp->who_new);
 	case OFFLINE:
-	    return wp->who_name;
+	    return strsave(wp->who_name);
 	default:
 	    break;
 	}
@@ -569,27 +555,27 @@ who_info(ptr, c, wbuf, wbufsiz)
     case 'a':
 	switch (wp->who_status & STMASK) {
 	case ONLINE:
-	    return CGETS(26, 9, "logged on");
+	    return strsave(CGETS(26, 9, "logged on"));
 	case OFFLINE:
-	    return CGETS(26, 10, "logged off");
+	    return strsave(CGETS(26, 10, "logged off"));
 	case CHANGED:
-	    xsnprintf(wbuf, wbufsiz, CGETS(26, 11, "replaced %s on"),
-		      wp->who_name);
-	    return wbuf;
+	    return xasprintf(CGETS(26, 11, "replaced %s on"), wp->who_name);
 	default:
 	    break;
 	}
 	break;
 
-#ifdef HAVE_STRUCT_UTMP_UT_HOST
+#ifdef UTHOSTLEN
     case 'm':
 	if (wp->who_host[0] == '\0')
-	    return CGETS(26, 12, "local");
+	    return strsave(CGETS(26, 12, "local"));
 	else {
+	    pb = wp->who_host;
+	    wbuf = xmalloc(strlen(pb) + 1);
+	    wb = wbuf;
 	    /* the ':' stuff is for <host>:<display>.<screen> */
-	    for (pb = wp->who_host,
-		flg = isdigit((unsigned char)*pb) ? '\0' : '.';
-		*pb != '\0' && (*pb != flg || ((pb = strchr(pb, ':')) != 0));
+	    for (flg = isdigit((unsigned char)*pb) ? '\0' : '.';
+		 *pb != '\0' && (*pb != flg || ((pb = strchr(pb, ':')) != 0));
 		 pb++) {
 		if (*pb == ':')
 		    flg = '\0';
@@ -602,20 +588,24 @@ who_info(ptr, c, wbuf, wbufsiz)
 
     case 'M':
 	if (wp->who_host[0] == '\0')
-	    return CGETS(26, 12, "local");
+	    return strsave(CGETS(26, 12, "local"));
 	else {
-	    for (pb = wp->who_host; *pb != '\0'; pb++)
+	    pb = wp->who_host;
+	    wbuf = xmalloc(strlen(pb) + 1);
+	    wb = wbuf;
+	    for (; *pb != '\0'; pb++)
 		*wb++ = isupper((unsigned char)*pb) ?
 		    tolower((unsigned char)*pb) : *pb;
 	    *wb = '\0';
 	    return wbuf;
 	}
-#endif /* HAVE_STRUCT_UTMP_UT_HOST */
+#endif /* UTHOSTLEN */
 
     case 'l':
-	return wp->who_tty;
+	return strsave(wp->who_tty);
 
     default:
+	wbuf = xmalloc(3);
 	wbuf[0] = '%';
 	wbuf[1] = (char) c;
 	wbuf[2] = '\0';
@@ -626,9 +616,7 @@ who_info(ptr, c, wbuf, wbufsiz)
 
 void
 /*ARGSUSED*/
-dolog(v, c)
-Char **v;
-struct command *c;
+dolog(Char **v, struct command *c)
 {
     struct who *wp;
     struct varent *vp;
@@ -646,15 +634,15 @@ struct command *c;
     }
 }
 
-# ifdef HAVE_STRUCT_UTMP_UT_HOST
+# ifdef UTHOSTLEN
 size_t
-utmphostsize()
+utmphostsize(void)
 {
     return UTHOSTLEN;
 }
 
 char *
-utmphost()
+utmphost(void)
 {
     char *tty = short2str(varval(STRtty));
     struct who *wp;
@@ -670,12 +658,11 @@ utmphost()
     resetwatch();
     return host;
 }
-# endif /* HAVE_STRUCT_UTMP_UT_HOST */
+# endif /* UTHOSTLEN */
 
 #ifdef WINNT_NATIVE
-void add_to_who_list(name, mach_nm)
-    char *name;
-    char *mach_nm;
+void
+add_to_who_list(char *name, char *mach_nm)
 {
 
     struct who *wp, *wpnew;
@@ -706,7 +693,7 @@ void add_to_who_list(name, mach_nm)
 	}
     }
     else {
-	wpnew = (struct who *) xcalloc(1, sizeof *wpnew);
+	wpnew = xcalloc(1, sizeof *wpnew);
 	(void) strncpy(wpnew->who_tty, mach_nm, UTLINLEN);
 	wpnew->who_time = 0;
 	if (*name == '\0')
